/* tkWWW.c Adds World Wide Web commands to a Tcl interpreter
** ===============
**
** Authors:
** Joseph Wang, Department of Astronomy, University of Texas at Austin
** (joe@astro.as.utexas.edu)
**
** History:
** 25 Jul 1992 Version 0.1 (JW)
**             Initial release of tkWWW
**
** 06 Aug 1992 Version 0.2 (JW)
**             Made procedures Private
**             Added comments to conform with WWW Style Guide
**             Added HtPing
**             Added ability to get source
**             Separated HtGoto into HtLoad and HtParseName
**             History recording is now not automatic
**
** 01 Sep 1992 Version 0.3 (JW)
**             Add argument count check for HtLoad to prevent 
**                segmentation violation
**             Removed HtCd (duplicates cd in standard tcl)
**             Removed HtPing
**             Removed HtHistoryList
**             Consolidated HtNext and HtPrevious into HtHistoryMoveBy
**             Removed unneeded global variables
**
** 18 Oct 1992 Version 0.4 (JW)
**             Added GetReference
**             Moved history handling to frontend
**             Added version retreval
**             Removed HtFind
**             Removed HtHistoryRecord
**             Moved reference_mark to GridText.c
**             Removed remaining globals
**             Changed HtGetText to generate commands
**
** 31 Jan 1993 Version 0.5 (JW)
**             Send PARSE_ALL to HtParse
**
** 20 Feb 1993 Version 0.6 (JW)
**             Added HtUncache command
**             Added HtGetIsIndex command
**             HtGetReference returns list of name and href
**             Added HtGetNumberFromTag command
**
** 14 Mar 1993 Version 0.7 (JW)
**             Switch to libWWW 2.0
**             Add internal links
**             Add HtSetSuffix
**             Add HtAddAnchor
**
** Copyright:
** Copyright (C) 1992 Joseph Wang
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Include c stuff
** ---------------
*/

#include <assert.h>
#include <stdio.h>
#include <ctype.h>

#ifdef __STDC__
#include <sys/types.h>
#include <stdlib.h>
#endif 

/* Include World Wide Web stuff
** ----------------------------
*/

#include <HTAnchor.h>
#include <HTParse.h>
#include <HTAccess.h>
#include <HTML.h>
#include <HText.h>
#include <HTTCP.h>
#include <HTStream.h>
#include <HTFile.h>
#include <HTFWriter.h>
#include <GridText.h>
#include <HTAtom.h>
#include <HTPlain.h>
#include <HTMLGen.h>
#include <HTFormat.h>
#include <HTMIME.h>
#include <HTInit.h>

/* Include tcl include header
** -------------------------
*/
#include <tcl.h>

/* Some Global Variables
** ---------------------
*/

PUBLIC  int  HTScreenWidth = 80;	/* By default */
PUBLIC  int  HTScreenHeight = 99999;	/* Undefined */
EXTERN HTPresentation* default_presentation;
					   
/* Macro to check arguments
** ------------------------
** Assumes interpreter is in interp and number of arguments is in argc
*/

#define HtCheckArgc(min,max,function_name) \
    if ((argc < (min)) || (argc > (max))) { \
       Tcl_AppendResult(interp, (function_name), \
			": Incorrect number of arguments" , NULL); \
       return (TCL_ERROR); \
       }

/* Procedures called by tkWWW functions
** ------------------------------------
*/

PRIVATE int HtGetAddressAndTitleCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc; 
     char **argv;
{
  char *address = HTAnchor_address((HTAnchor *) HTMainAnchor);
  Tcl_AppendElement(interp, address, 0);
  Tcl_AppendElement(interp, (char *) HTAnchor_title(HTMainAnchor), 0);
  free (address);
  return (TCL_OK);
}

PRIVATE int HtGetIsIndexCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc; 
     char **argv;
{   
  Tcl_SetResult(interp, 
		(HTAnchor_isIndex(HTAnchor_parent((HTAnchor *)HTMainAnchor)) 
		 ? "1" : "0"),
		TCL_STATIC);
  return (TCL_OK);
}

PRIVATE int HtGetNumberFromTagCmd(dummy, interp, argc, argv) 
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HTChildAnchor *source;
  int i;
  char buffer[255];

  HtCheckArgc(2, 2, "HtGetNumberFromTag");
  for (i= 1; i <= HText_sourceAnchors(HTMainText) ; i++) {
    source = HText_childNumber(HTMainText, i);
    if (source && source->tag && !strcmp(source->tag, argv[1])) {
      sprintf(buffer, "%d", i);
      Tcl_SetResult(interp, buffer, TCL_VOLATILE);
      return (TCL_OK);
    } 
  }

  return (TCL_OK);
}

PRIVATE int HtGetReferenceCmd(dummy, interp, argc, argv) 
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  int reference_num;
  HTChildAnchor *source;
  CONST char *address;

  if (HTMainText) {
    HtCheckArgc(2, 2, "HtGetReference");
    reference_num = atoi(argv[1]);
    if ((reference_num < 1) ||
	(reference_num > HText_sourceAnchors(HTMainText))) {
      Tcl_SetResult(interp, "Reference out of range", TCL_STATIC);
      return (TCL_ERROR);
    }
    
    source = HText_childNumber(HTMainText, reference_num);
    if (!source) {
      Tcl_SetResult(interp, "No child anchor", TCL_STATIC);
      return (TCL_ERROR);
    }
  } else {
    Tcl_SetResult(interp, "No main text", TCL_STATIC);
    return (TCL_ERROR);
  }

  Tcl_AppendElement(interp, source->tag, 0);

  address = HText_getAnchorHref(HTMainText, reference_num);
  Tcl_AppendElement(interp, address, 0);

  return (TCL_OK);
}


PRIVATE int HtGetSourceCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc; 
     char **argv;
{   
  Tcl_SetResult(interp, (char *) HText_getSource(HTMainText), TCL_STATIC);
  return (TCL_OK);
} 


PRIVATE int HtGetTextCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc; 
     char **argv;
{   
  HText *text = HTMainText;
  if (!text) {
    Tcl_SetResult (interp, "No main text", TCL_STATIC);
    return (TCL_ERROR);
  }
  Tcl_SetResult(interp, text->output->data, TCL_STATIC);
  return (TCL_OK);
} 

PRIVATE int HtLoadCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HtCheckArgc(2,2, "HtLoad");
  if (!HTLoadAbsolute(argv[1])) {
    Tcl_SetResult(interp, "Page not found: ", TCL_STATIC);
    Tcl_AppendResult(interp, argv[1], NULL);
    return (TCL_ERROR);
  }
  return (TCL_OK);
}

PRIVATE int HtParseNameCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  char *current_address = NULL;
  char *ref;
  
  HtCheckArgc(2,2,"HtParseName");
  if (HTMainAnchor) {
    current_address = HTAnchor_address((HTAnchor *) HTMainAnchor);
  } else {
    current_address = NULL;
    StrAllocCopy(current_address, "file:");

/* The following mess is to get the current working directory in a manner
 * that will work on most platforms 
 */

#ifndef MAXPATHLEN
#define NO_GETWD		/* Assume no  getwd() if no MAXPATHLEN */
#endif

#ifdef NO_GETWD  		/* No getwd() on this machine */
#ifdef HAS_GETCWD		/* System V variant SIGN CHANGED TBL 921006 !! */

    {
      char wd[1024];			/*!! Arbitrary*/
      extern char * getcwd();
      char * result = getcwd(wd, sizeof(wd)); 
      if (result) {

#ifdef vms  /* convert directory name to Unix-style syntax */
	char * disk = strchr (wd, ':');
	char * dir = strchr (wd, '[');
	if (disk) {
	  *disk = '\0';
	  StrAllocCat (current_address, "/");  /* needs delimiter */
	  StrAllocCat (current_address, wd);
	}
	if (dir) {
	  char *p;
	  *dir = '/';  /* Convert leading '[' */
	  for (p = dir ; *p != ']'; ++p)
	    if (*p == '.') *p = '/';
	  *p = '\0';  /* Cut on final ']' */
	  StrAllocCat (current_address, dir);
	}
#else  /* not VMS */
	StrAllocCat (current_address, wd);
#endif  /* not VMS */
      } else {
	fprintf(stderr,
		"HTBrowse: Can't read working directory (getcwd).\n");
      }
    }  /* end if good getcwd result */
	
#else   /* has NO getcwd */

    if (TRACE) 
      fprintf(stderr,
	      "HTBrowse: This platform does not support getwd() or getcwd()\n");
#endif	/* has no getcwd */

#else   /* has getwd */
    {
      char wd[MAXPATHLEN];
      extern char * getwd();
      char * result = getwd(wd);
      if (result) {
	StrAllocCat(current_address, wd);
      } else {
	fprintf(stderr, "HTBrowse: Can't read working directory.\n");
      }
    }
#endif
		
#ifdef vms
    StrAllocCat(current_address, "default.html");
#else
    StrAllocCat(current_address, "/default.html");
#endif
  }

  ref = HTParse(argv[1], current_address, PARSE_ALL);  
  Tcl_SetResult(interp, ref, TCL_DYNAMIC);
  free(current_address);
  
  return (TCL_OK);
}

PRIVATE int HtUncacheCmd (dummy, interp, argc, argv) 
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  HTParentAnchor *parent_anchor;
  HTAnchor *anchor;
  HText *document;
  HtCheckArgc(2, 2, "HtUncache");

  anchor = HTAnchor_findAddress(argv[1]);
  if (anchor) {
    parent_anchor = HTAnchor_parent(anchor);
    document = (HText *)HTAnchor_document(parent_anchor);
    if (document) {
      if (document == HTMainText)
	HTMainText = NULL;
      HText_free(document);
    }
  }
  return (TCL_OK);
}

  
/* HtVersion
** ------------
** returns the version of the tkWWW server
*/

PRIVATE int HtVersionCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  
  Tcl_SetResult(interp, SERVER_VERSION, TCL_STATIC);
  return (TCL_OK);
}

PUBLIC void HTTkSetOutputFile(format, anchor, fnam)
     HTFormat format;
     HTParentAnchor *anchor;
     CONST char *fnam;
 {
  HText *text = HText_new(anchor);
  HTChunkPuts(text->output, "tkWWWConfigDisplayFile ");
  HTChunkPuts(text->output, HTAtom_name(format));
  HTChunkPuts(text->output, " ");
  HTChunkPuts(text->output, fnam);
  HTChunkTerminate(text->output);
}

PRIVATE HTStream * HTTkDisplay(pres, anchor, sink) 
     HTPresentation *pres;
     HTParentAnchor *anchor;
     HTStream *sink;

{
  char *fnam;
  CONST char * suffix;
  HTStream* this;
  FILE *fp;
  /* Save the file under a suitably suffixed name */

  suffix = HTFileSuffix(pres->rep);

  fnam = (char *)malloc (L_tmpnam + 16 + strlen(suffix));
  tmpnam (fnam);
  if (suffix) strcat(fnam, suffix);
    
  fp = fopen (fnam, "w");
  if (!fp) {
    HTAlert("Can't open temporary file!");
    free(fnam);
    return NULL;
  }

  this = HTFWriter_new(fp);
  HTTkSetOutputFile(pres->rep, anchor, fnam);
  free (fnam);
  return this;
}

PUBLIC BOOL HTTkUseInPlace(format_in, format_out)
     HTFormat format_in;
     HTFormat format_out;
{
  int n = HTList_count(HTPresentations);
  int i;
  HTPresentation *pres;
  for (i=0; i<n; i++) {
    pres = HTList_objectAt(HTPresentations, i);
    if (pres->rep == format_in && pres->rep_out == format_out) 
      return (pres->converter == HTTkDisplay) ? TRUE : FALSE;
  }
  if (default_presentation && 
      default_presentation->converter == HTTkDisplay)
    return TRUE;
  return FALSE;

}

PRIVATE int HtSetSuffixCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  float priority;
  HtCheckArgc(3, 4, "HtSetSuffix");
  priority  = (argc==3) ? 1.0 : atof(argv[3]);
  HTSetSuffix(argv[1], argv[2], priority);
  return (TCL_OK);
}

PRIVATE int HtAddAnchorCmd(dummy, interp, argc, argv) 
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  int anchor_number;
  char *anchor_name = NULL, *anchor_address = NULL, buffer[255];
  HTChildAnchor *source;
  HtCheckArgc(2, 3, "HtAddAnchorCmd");
  
  if (HTMainAnchor) {
    anchor_name = argv[1];
    if (argc == 3 && (*argv[2] != '\00'))
      anchor_address = argv[2];

    source = 
      HTAnchor_findChildAndLink(HTMainAnchor,                /* parent */
				anchor_name,	/* Tag */
				anchor_address,	/* Addresss */
				0);
    anchor_number = HText_addAnchor(HTMainText, source);
    HText_setAnchorHref(HTMainText, anchor_address);
    
    sprintf(buffer, "%d", anchor_number);
    Tcl_SetResult(interp, buffer, TCL_VOLATILE);
    return (TCL_OK);
  } else {
    return (TCL_ERROR);
  }
}

/* Add tkWWW commands to a tcl interpreter init_tkWWW(interp)
** -------------
*/

void init_tkWWW(interp) 
     Tcl_Interp *interp;
{
  Tcl_CreateCommand(interp, "HtAddAnchor", HtAddAnchorCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtGetAddressAndTitle", HtGetAddressAndTitleCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtGetIsIndex", HtGetIsIndexCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtGetNumberFromTag", HtGetNumberFromTagCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtGetReference", HtGetReferenceCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtGetSource", HtGetSourceCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtGetText", HtGetTextCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtSetSuffix", HtSetSuffixCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtLoad", HtLoadCmd, 
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtParseName", HtParseNameCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtUncache", HtUncacheCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  Tcl_CreateCommand(interp, "HtVersion", HtVersionCmd,
		    (ClientData) NULL, (void (*)()) NULL);
  HTSetConversion("www/mime",  "*",           HTMIMEConvert, 	1.0, 0.0, 0.0);
  HTSetConversion("text/html", "text/x-c",    HTMLToC, 	        0.5, 0.0, 0.0);
  HTSetConversion("text/html", "text/plain",  HTMLToPlain, 	0.5, 0.0, 0.0);
  HTSetConversion("text/html", "www/present", HTMLPresent, 	1.0, 0.0, 0.0);
  HTSetConversion("text/plain", "text/html",  HTPlainToHTML,	1.0, 0.0, 0.0);
  HTSetConversion("*",          "www/present",HTTkDisplay,      0.5, 0.0, 0.0);
}
